<?php

/**
 * @file
 * Database interface code for automatic slave selection depending on type of
 * query.
 *
 * @todo Ensure default method arguments will work if they differ from what's
 *       defined in class Database.
 */

/**
 * @ingroup database
 * @{
 */

/**
 * Specific auto slave implementation of DatabaseConnection.
 */
class DatabaseConnection_autoslave extends DatabaseConnection {

  /**
   * List of tables that should use "master" as target
   */
  private $__tables = array();

  /**
   * Force queries to master
   */
  private $__force_master = FALSE;

  static protected $__master = 'master';
  static protected $__slave = 'slave';

  /**
   * Constructor.
   */
  public function __construct(array $connection_options = array()) {
    // Initialize force master tables
    $this->__tables = variable_get('autoslave_tables', array());
    if (!empty($this->__tables)) {
      $this->__tables = array_combine($this->__tables, $this->__tables);
    }

    // Has master been forced before Database bootstrapping?
    if (!empty($_GLOBALS['autoslave_pre_bootstrap_force_master'])) {
      $this->forceMaster(1);
    }
  }

  /**
   * Dispatch all methods defined in class PDO to the appropiate backend
   */
  public function beginTransaction() { return $this->getMaster()->beginTransaction(); }
  public function errorCode() { return $this->getMaster()->errorCode(); }
  public function errorInfo() { return $this->getMaster()->errorInfo(); }
  public function exec($statement) { return $this->getMaster()->exec($statement); }
  public function getAttribute($attribute) { return $this->getMaster()->getAttribute($attribute); }
  static public function getAvailableDrivers() { $drivers = $this->getMaster()->getAvailableDrivers(); $drivers[] = 'autoslave'; return $drivers; }
  public function lastInsertId($name = NULL) { return $this->getMaster()->lastInsertId($name); }
  public function quote() { return $this->getMaster()->quote(); }
  public function setAttribute($attribute, $value) { return $this->getMaster()->setAttribute($attribute, $value); }

  /**
   * Dispatch all methods defined in class Database to the appropiate backend
   */
  protected function defaultOptions() { return $this->getMaster()->defaultOptions(); }
  public function getConnectionOptions() { return $this->getMaster()->getConnectionOptions(); }
  protected function setPrefix($prefix) { return $this->getMaster()->setPrefix($prefix); }
  public function prefixTables($sql) { return $this->getMaster()->prefixTables($sql); }
  public function tablePrefix($table = 'default') { return $this->getMaster()->tablePrefix($table); }
  public function prepareQuery($query) { return $this->getMaster()->prepareQuery($query); }
  public function setTarget($target = NULL) { return $this->getMaster()->setTarget($target); }
  public function getTarget() { return $this->getMaster()->getTarget(); }
  public function setKey($key) { return $this->getMaster()->setKey($key); }
  public function getKey() { return $this->getMaster()->getKet(); }
  public function setLogger(DatabaseLog $logger) { return $this->getMaster()->setLogger($logger); }
  public function getLogger() { return $this->getMaster()->getLogger(); }
  public function makeSequenceName($table, $field) { return $this->getMaster()->makeSequenceName($table, $field); }
  public function makeComment($comments) { return $this->getMaster()->makeComment($comments); }
  protected function filterComment($comment = '') { return $this->getMaster()->filterComment($comment); }
  protected function expandArguments(&$query, &$args) { return $this->getMaster()->expandArguments($query, $args); }
  public function escapeTable($table) { return $this->getMaster()->escapeTable($table); }
  public function escapeField($field) { return $this->getMaster()->escapeField($field); }
  public function escapeAlias($field) { return $this->getMaster()->escapeAlias($field); }
  public function escapeLike($string) { return $this->getMaster()->escapeLike($string); }
  public function inTransaction() { return $this->getMaster()->inTransaction(); }
  public function transactionDepth() { return $this->getMaster()->transactionDepth(); }
  public function rollback($savepoint_name = 'drupal_transaction') { return $this->getMaster()->rollback($savepoint_name); }
  public function pushTransaction($name) { return $this->getMaster()->pushTransaction($name); }
  public function popTransaction($name) { return $this->getMaster()->popTransaction($name); }
  protected function popCommittableTransactions() { return $this->getMaster()->popCommittableTransactions(); }
  protected function generateTemporaryTableName() { return $this->getMaster()->generateTemporaryTableName(); }
  public function supportsTransactions() { return $this->getMaster()->supportsTransactions(); }
  public function supportsTransactionalDDL() { return $this->getMaster()->supportsTransactions(); }
  function commit() { return $this->getMaster()->commit(); }

  /**
   * The following methods are absolutely necessary to overload manually (at least for MySQL)
   */

  #public function getDriverClass($class, array $files = array(), $use_autoload = FALSE) { return $this->getMaster()->getDriverClass($class, $files, $use_autoload); }
  public function version() { return $this->getMaster()->version(); }
  public function schema() { return $this->getMaster()->schema(); }
  public function driver() { return 'autoslave'; }
  public function databaseType() { return $this->getMaster()->databaseType(); }
  public function query($query, array $args = array(), $options = array()) { return Database::getConnection($this->deriveTargetFromQuery($query))->query($query, $args, $options); }
  public function queryRange($query, $from, $count, array $args = array(), array $options = array()) { return Database::getConnection($this->deriveTargetFromQuery($query))->queryRange($query, $from, $count, $args, $options); }
  public function queryTemporary($query, array $args = array(), array $options = array()) { return $this->getMaster()->queryTemporary($query, $args, $options); }
  public function mapConditionOperator($operator) { return $this->getMaster()->mapConditionOperator($operator); }
  public function nextId($existing_id = 0) { return $this->getMaster()->nextId($existing_id); }

  public function select($table, $alias = NULL, array $options = array()) {
    return Database::getConnection($this->getTargetForTable($table))->select($table, $alias, $options);
  }
  public function insert($table, $options = array()) {
    $this->addAffectedTable($table);
    return $this->getMaster()->insert($table, $options);
  }

  public function merge($table, $options = array()) {
    $this->addAffectedTable($table);
    return $this->getMaster()->merge($table, $options);
  }

  public function update($table, $options = array()) {
    $this->addAffectedTable($table);
    return $this->getMaster()->update($table, $options);
  }

  public function delete($table, $options = array()) {
    $this->addAffectedTable($table);
    return $this->getMaster()->delete($table, $options);
  }

  public function truncate($table, $options = array()) {
    $this->addAffectedTable($table);
    return $this->getMaster()->truncate($table, $options);
  }

  public function startTransaction($name = '') {
    return new DatabaseTransaction_autoslave($this, $name);
  }


  /**
   * Magic methods. Supports propeties/functions/methods not defined by the Database class
   */
  public static function __callStatic($method, $args) {
    return call_user_func_array(array(Database::getConnection(self::$__master)->get_class_name(), $method), $args);
  }

  public function __call($method, $args) {
    return call_user_func_array(array($this->getMaster(), $method), $args);
  }

  public function __get($key) {
    return $this->getMaster()->$key;
  }

  public function __set($key, $value) {
    $this->getMaster()->$key = $value;
  }






  /**
   * Internal autoslave functions
   */

  /**
   * Determine the which target to use based on a query.
   */
  private function deriveTargetFromQuery($query) {
    #print "$query : " . $this->forceMaster() . "<br>";
    $is_write_query = preg_match('/^\s*('.
      'UPDATE|INSERT|REPLACE|DELETE|'.
      'ALTER|CREATE|DROP|TRUNCATE|RENAME|'.
      'BEGIN|START\s+TRANSACTION|COMMIT|ROLLBACK|'.
      'RELEASE|SAVEPOINT|'.
      '(.*FOR UPDATE$)|(.*LOCK IN SHARE MODE$)'.
    ')\b/i', $query);

    // Find all tables used in the query.
    preg_match_all('@\{(\w+)\}@', $query, $matches);
    $tables = $matches[1];

    if ($is_write_query) {
      // Even if forceMaster() is true, we still need to tag tables that have been written to,
      // in case we go back to forceMaster() false later on.
      $this->addAffectedTables($tables);
      return self::$__master;
    }
    elseif ($this->forceMaster()) {
      return self::$__master;
    }
    else {
      return array_intersect($tables, $this->__tables) ? self::$__master : self::$__slave;
    }

  }

  /**
   * Get the target for a table
   *
   * @param $table
   *   Table to get target for
   * @return string
   *   Target
   */
  private function getTargetForTable($table) {
    return $this->forceMaster() || isset($this->__tables[$table]) ? self::$__master : self::$__slave;      
  }

  /**
   * Get/set force master counter
   */
  public function forceMaster($force_master = NULL) {
    if (isset($force_master)) {
      $this->__force_master += $force_master;
    }
    return $this->__force_master;
  }

  /**
   * Get the master connection
   */
  public function getMaster() {
    return Database::getConnection(self::$__master);
  }

  /**
   * Get list of affected tables.
   */
  public function getAffectedTables() {
    return $this->__tables;
  }

  /**
   * Add tables to list of affected tables.
   */
  public function addAffectedTables($tables, $update_session = TRUE) {
    $this->__tables = array_unique(array_merge($this->__tables, $tables));
    if ($update_session) {
      $_SESSION['autoslave_affected_tables'] = $this->__tables;
      $_SESSION['autoslave_affected_tables_timestamp'] = time();
    }
    return $this->__tables;
  }

  public function addAffectedTable($table, $update_session = TRUE) {
    return $this->addAffectedTables(array($table => $table), $update_session);
  }
}

/**
 * Handle force master for transactions
 */
class DatabaseTransaction_autoslave extends DatabaseTransaction {
  public $autoslave_connection;
  protected $parent;

  public function __construct(DatabaseConnection $connection, $name = NULL) {
    $this->autoslave_connection = $connection;
    $this->autoslave_connection->forceMaster(1);
    $this->parent = $this->autoslave_connection->getMaster()->startTransaction($name);
  }

  public function __destruct() {
    unset($this->parent);
    $this->autoslave_connection->forceMaster(-1);
  }

  public function name() { return $this->parent->name(); }
  public function rollback() { return $this->parent->rollback(); }
}

